@@ -448,7 +448,7 @@ class ElucidatedImagen(nn.Module):
                 if has_inpainting and not (is_last_resample_step or is_last_timestep):
                     # renoise in repaint and then resample
                     repaint_noise = torch.randn(shape, device = self.device)
-                    images = images + (sigma_hat - sigma_next) * repaint_noise
+                    images = images + (sigma - sigma_next) * repaint_noise
 
         images = images.clamp(-1., 1.)
 
