@@ -170,133 +170,6 @@ def prob_mask_like(shape, prob, device):
     else:
         return torch.zeros(shape, device = device).float().uniform_(0, 1) < prob
 
-# gaussian diffusion helper functions
-
-def extract(a, t, x_shape):
-    b, *_ = t.shape
-    out = a.gather(-1, t)
-    return out.reshape(b, *((1,) * (len(x_shape) - 1)))
-
-def cosine_beta_schedule(timesteps, s = 0.008, thres = 0.999):
-    """
-    cosine schedule
-    as proposed in https://openreview.net/forum?id=-NEXDKk8gZ
-    """
-    steps = timesteps + 1
-    x = torch.linspace(0, timesteps, steps, dtype = torch.float64)
-    alphas_cumprod = torch.cos(((x / timesteps) + s) / (1 + s) * torch.pi * 0.5) ** 2
-    alphas_cumprod = alphas_cumprod / alphas_cumprod[0]
-    betas = 1 - (alphas_cumprod[1:] / alphas_cumprod[:-1])
-    return torch.clip(betas, 0, thres)
-
-def linear_beta_schedule(timesteps):
-    scale = 1000 / timesteps
-    beta_start = scale * 0.0001
-    beta_end = scale * 0.02
-    return torch.linspace(beta_start, beta_end, timesteps, dtype = torch.float64)
-
-class GaussianDiffusion(nn.Module):
-    def __init__(
-        self,
-        *,
-        noise_schedule,
-        timesteps
-    ):
-        super().__init__()
-
-        if noise_schedule == "cosine":
-            betas = cosine_beta_schedule(timesteps)
-        elif noise_schedule == "linear":
-            betas = linear_beta_schedule(timesteps)
-        else:
-            raise NotImplementedError()
-
-        alphas = 1. - betas
-        alphas_cumprod = torch.cumprod(alphas, axis = 0)
-        alphas_cumprod_prev = F.pad(alphas_cumprod[:-1], (1, 0), value = 1.)
-
-        timesteps, = betas.shape
-        self.num_timesteps = int(timesteps)
-
-        # register buffer helper function to cast double back to float
-
-        register_buffer = lambda name, val: self.register_buffer(name, val.to(torch.float32), persistent = False)
-
-        register_buffer('betas', betas)
-        register_buffer('alphas_cumprod', alphas_cumprod)
-        register_buffer('alphas_cumprod_prev', alphas_cumprod_prev)
-
-        # calculations for diffusion q(x_t | x_{t-1}) and others
-
-        register_buffer('sqrt_alphas_cumprod', torch.sqrt(alphas_cumprod))
-        register_buffer('sqrt_one_minus_alphas_cumprod', torch.sqrt(1. - alphas_cumprod))
-        register_buffer('log_one_minus_alphas_cumprod', torch.log(1. - alphas_cumprod))
-        register_buffer('sqrt_recip_alphas_cumprod', torch.sqrt(1. / alphas_cumprod))
-        register_buffer('sqrt_recipm1_alphas_cumprod', torch.sqrt(1. / alphas_cumprod - 1))
-
-        # calculations for posterior q(x_{t-1} | x_t, x_0)
-
-        posterior_variance = betas * (1. - alphas_cumprod_prev) / (1. - alphas_cumprod)
-
-        # above: equal to 1. / (1. / (1. - alpha_cumprod_tm1) + alpha_t / beta_t)
-
-        register_buffer('posterior_variance', posterior_variance)
-
-        # below: log calculation clipped because the posterior variance is 0 at the beginning of the diffusion chain
-
-        register_buffer('posterior_log_variance_clipped', log(posterior_variance, eps = 1e-20))
-        register_buffer('posterior_mean_coef1', betas * torch.sqrt(alphas_cumprod_prev) / (1. - alphas_cumprod))
-        register_buffer('posterior_mean_coef2', (1. - alphas_cumprod_prev) * torch.sqrt(alphas) / (1. - alphas_cumprod))
-
-    def get_times(self, batch_size, noise_level, *, device):
-        return torch.full((batch_size,), int(self.num_timesteps * noise_level), device = device, dtype = torch.long)
-
-    def sample_random_times(self, batch_size, *, device):
-        return torch.randint(0, self.num_timesteps, (batch_size,), device = device, dtype = torch.long)
-
-    def get_condition(self, times):
-        return times
-
-    def get_sampling_timesteps(self, batch, *, device):
-        time_transitions = []
-
-        for i in reversed(range(self.num_timesteps)):
-            time_transitions.append((torch.full((batch,), i, device = device, dtype = torch.long), None))
-
-        return time_transitions
-
-    def q_posterior(self, x_start, x_t, t, **kwargs):
-        posterior_mean = (
-            extract(self.posterior_mean_coef1, t, x_t.shape) * x_start +
-            extract(self.posterior_mean_coef2, t, x_t.shape) * x_t
-        )
-        posterior_variance = extract(self.posterior_variance, t, x_t.shape)
-        posterior_log_variance_clipped = extract(self.posterior_log_variance_clipped, t, x_t.shape)
-        return posterior_mean, posterior_variance, posterior_log_variance_clipped
-
-    def q_sample(self, x_start, t, noise = None):
-        if isinstance(t, int):
-            batch = x_start.shape[0]
-            t = torch.full((batch,), t, device = self.betas.device)
-
-        noise = default(noise, lambda: torch.randn_like(x_start))
-
-        noised = (
-            extract(self.sqrt_alphas_cumprod, t, x_start.shape) * x_start +
-            extract(self.sqrt_one_minus_alphas_cumprod, t, x_start.shape) * noise
-        )
-
-        alphas_cumprod = extract(self.alphas_cumprod, t, t.shape)
-        log_snr = -log(1. / alphas_cumprod - 1)
-
-        return noised, log_snr
-
-    def predict_start_from_noise(self, x_t, t, noise):
-        return (
-            extract(self.sqrt_recip_alphas_cumprod, t, x_t.shape) * x_t -
-            extract(self.sqrt_recipm1_alphas_cumprod, t, x_t.shape) * noise
-        )
-
 # gaussian diffusion with continuous time helper functions and classes
 # large part of this was thanks to @crowsonkb at https://github.com/crowsonkb/v-diffusion-jax/blob/master/diffusion/utils.py
 
@@ -371,6 +244,28 @@ class GaussianDiffusionContinuousTimes(nn.Module):
         alpha, sigma =  log_snr_to_alpha_sigma(log_snr_padded_dim)
         return alpha * x_start + sigma * noise, log_snr
 
+    def q_sample_from_to(self, x_from, from_t, to_t, noise = None):
+        shape, device, dtype = x_from.shape, x_from.device, x_from.dtype
+        batch = shape[0]
+
+        if isinstance(from_t, float):
+            from_t = torch.full((batch,), from_t, device = device, dtype = dtype)
+
+        if isinstance(to_t, float):
+            to_t = torch.full((batch,), to_t, device = device, dtype = dtype)
+
+        noise = default(noise, lambda: torch.randn_like(x_from))
+
+        log_snr = self.log_snr(from_t)
+        log_snr_padded_dim = right_pad_dims_to(x_from, log_snr)
+        alpha, sigma =  log_snr_to_alpha_sigma(log_snr_padded_dim)
+
+        log_snr_to = self.log_snr(from_t)
+        log_snr_padded_dim_to = right_pad_dims_to(x_from, log_snr_to)
+        alpha_to, sigma_to =  log_snr_to_alpha_sigma(log_snr_padded_dim_to)
+
+        return x_from * (alpha_to / alpha) + noise * (sigma_to * alpha - sigma * alpha_to) / alpha
+
     def predict_start_from_noise(self, x_t, t, noise):
         log_snr = self.log_snr(t)
         log_snr = right_pad_dims_to(x_t, log_snr)
@@ -1765,7 +1660,6 @@ class Imagen(nn.Module):
         per_sample_random_aug_noise_level = False,  # unclear when conditioning on augmentation noise level, whether each batch element receives a random aug noise value - turning off due to @marunine's find
         condition_on_text = True,
         auto_normalize_img = True,                  # whether to take care of normalizing the image from [0, 1] to [-1, 1] and back automatically - you can turn this off if you want to pass in the [-1, 1] ranged image yourself from the dataloader
-        continuous_times = True,
         p2_loss_weight_gamma = 0.5,                 # p2 loss weight, from https://arxiv.org/abs/2204.00227 - 0 is equivalent to weight of 1 across time
         p2_loss_weight_k = 1,
         dynamic_thresholding = True,
@@ -1815,7 +1709,7 @@ class Imagen(nn.Module):
 
         # construct noise schedulers
 
-        noise_scheduler_klass = GaussianDiffusion if not continuous_times else GaussianDiffusionContinuousTimes
+        noise_scheduler_klass = GaussianDiffusionContinuousTimes
         self.noise_schedulers = nn.ModuleList([])
 
         for timestep, noise_schedule in zip(timesteps, noise_schedules):
@@ -1859,7 +1753,7 @@ class Imagen(nn.Module):
                 text_embed_dim = self.text_embed_dim if self.condition_on_text else None,
                 channels = self.channels,
                 channels_out = self.channels,
-                learned_sinu_pos_emb = continuous_times
+                learned_sinu_pos_emb = True
             )
 
             self.unets.append(one_unet)
@@ -2053,6 +1947,7 @@ class Imagen(nn.Module):
         cond_images = None,
         inpaint_images = None,
         inpaint_masks = None,
+        inpaint_resample_times = 5,
         cond_scale = 1,
         pred_objective = 'noise',
         dynamic_threshold = True,
@@ -2067,6 +1962,7 @@ class Imagen(nn.Module):
         # prepare inpainting
 
         has_inpainting = exists(inpaint_images) and exists(inpaint_masks)
+        resample_times = inpaint_resample_times if has_inpainting else 1
 
         if has_inpainting:
             inpaint_images = self.normalize_img(inpaint_images)
@@ -2078,26 +1974,39 @@ class Imagen(nn.Module):
         timesteps = noise_scheduler.get_sampling_timesteps(batch, device = device)
 
         for times, times_next in tqdm(timesteps, desc = 'sampling loop time step', total = len(timesteps), disable = not use_tqdm):
+            is_last_timestep = times_next == 0
 
-            if has_inpainting:
-                noised_inpaint_images, _ = noise_scheduler.q_sample(inpaint_images, t = times)
-                img = img * ~inpaint_masks + noised_inpaint_images * inpaint_masks
+            for r in reversed(range(resample_times)):
+                is_last_resample_step = r == 0
 
-            img = self.p_sample(
-                unet,
-                img,
-                times,
-                t_next = times_next,
-                text_embeds = text_embeds,
-                text_mask = text_mask,
-                cond_images = cond_images,
-                cond_scale = cond_scale,
-                lowres_cond_img = lowres_cond_img,
-                lowres_noise_times = lowres_noise_times,
-                noise_scheduler = noise_scheduler,
-                pred_objective = pred_objective,
-                dynamic_threshold = dynamic_threshold
-            )
+                if has_inpainting:
+                    noised_inpaint_images, _ = noise_scheduler.q_sample(inpaint_images, t = times)
+                    img = img * ~inpaint_masks + noised_inpaint_images * inpaint_masks
+
+                img = self.p_sample(
+                    unet,
+                    img,
+                    times,
+                    t_next = times_next,
+                    text_embeds = text_embeds,
+                    text_mask = text_mask,
+                    cond_images = cond_images,
+                    cond_scale = cond_scale,
+                    lowres_cond_img = lowres_cond_img,
+                    lowres_noise_times = lowres_noise_times,
+                    noise_scheduler = noise_scheduler,
+                    pred_objective = pred_objective,
+                    dynamic_threshold = dynamic_threshold
+                )
+
+                if has_inpainting and not (is_last_resample_step or torch.all(is_last_timestep)):
+                    renoised_img = noise_scheduler.q_sample_from_to(img, times_next, times)
+
+                    img = torch.where(
+                        rearrange(is_last_timestep, 'b -> b 1 1 1'),
+                        img,
+                        renoised_img
+                    )
 
         img.clamp_(-1., 1.)
 
